import json

from taal import TranslationContextManager, translation_manager


class TypeTranslationContextManager(TranslationContextManager):
    """ TranslationContextManager for Kaiso types """

    context = "taal:kaiso_type"

    def __init__(self, manager, **kwargs):
        self.manager = manager

    def list_message_ids(self):
        manager = self.manager
        type_hierarchy = manager.get_type_hierarchy()
        return (type_[0] for type_ in type_hierarchy)

translation_manager.register(TypeTranslationContextManager)


class AttributeTranslationContextManager(TranslationContextManager):
    """ TranslationContextManager for Kaiso attributes """

    context = "taal:kaiso_attr"

    def __init__(self, manager, **kwargs):
        self.manager = manager

    @staticmethod
    def get_message_id(type_id, attr):
        return json.dumps(
            [type_id, attr.name]
        )

    def list_message_ids(self):
        manager = self.manager
        type_hierarchy = manager.get_type_hierarchy()
        for type_id, bases, attrs in type_hierarchy:
            for attr in attrs:
                yield self.get_message_id(
                    type_id, attr)

translation_manager.register(AttributeTranslationContextManager)



""" Taal: Translations for SQLAlchemy and Kaiso models

    Store and manage translations using SQLAlchemy
"""

from __future__ import absolute_import

from abc import ABCMeta, abstractmethod, abstractproperty

from sqlalchemy.orm import Session, aliased
from sqlalchemy.sql.expression import tuple_, and_, or_

from taal.exceptions import BindError


try:
    VERSION = __import__('pkg_resources').get_distribution('taal').version
except:  # pragma: no cover
    VERSION = 'unknown'


NULL = None  # for pep8


class TranslatableString(object):
    """
    Placeholder for a string to be translated

    Holds metadata, ``context`` and ``message_id``, and optionally
    a string ``pending_value``
    """

    def __init__(self, context=None, message_id=None, pending_value=None):
        self.context = context
        self.message_id = message_id
        self.pending_value = pending_value

    def __repr__(self):
        return "<TranslatableString: ({}, {}, {})>".format(
            self.context, self.message_id, self.pending_value)

    def __eq__(self, other):
        if not isinstance(other, TranslatableString):
            return False

        self_data = (self.context, self.message_id, self.pending_value)
        other_data = (other.context, other.message_id, other.pending_value)
        return self_data == other_data


class Translator(object):
    """
    Manage a particular set of translations

    Given a sqlalchemy session, a model to store translations, and
    a language, bind a translator to a(n other) sqlalchemy session
    and/or a kaiso manager to get translation magic

    In addition to native data types, attributes will also include
    instances of ``TranslatableString``. A translator may subsequently
    be passed "structured" data (dicts, lists, tuples) containing
    translatable strings and translate to a particular language
    """

    def __init__(self, model, session, language, debug_output=False):
        self.model = model
        self.session = session
        self.language = language
        self.debug_output = debug_output

    def bind(self, target):
        """ register e.g. a sqlalchey session or a kaiso manager """
        from taal.kaiso import manager

        if isinstance(target, Session):
            from taal.sqlalchemy.events import (
                register_translator, register_session)
            register_translator(target, self)
            register_session(target)
        elif isinstance(target, manager.Manager):
            manager.register_translator(target, self)
        else:
            raise BindError("Unknown target {}".format(target))

    def _get_debug_translation(self, translatable):
        return "[Translation missing ({}, {}, {})]".format(
            self.language, translatable.context, translatable.message_id)

    def _translate(self, translatable, cache):
        try:
            return cache[(translatable.context, translatable.message_id)]
        except KeyError:
            if self.debug_output:
                return self._get_debug_translation(translatable)
            return None

    def translate(self, translatable, cache=None):
        """
        Translate ``TranslatableString`` by looking up a translation

        can also take a 'structure' (currently lists, tuples, and dicts)
        and recursively translate any TranslatableStrings found.
        """

        if cache is None:
            cache = self._prepare_cache(translatable)

        if isinstance(translatable, TranslatableString):
            return self._translate(translatable, cache)
        elif isinstance(translatable, dict):
            return dict(
                (key, self.translate(val, cache))
                for key, val in translatable.iteritems()
            )
        elif isinstance(translatable, list):
            return list(
                self.translate(item, cache)
                for item in translatable)
        elif isinstance(translatable, tuple):
            return tuple(
                self.translate(item, cache)
                for item in translatable)

        else:
            return translatable

    def _prepare_cache(self, translatable):
        """
        Bulk load translations required to translate a translatable
        'structure'
        """
        translatables = self._collect_translatables(translatable)
        if not translatables:
            return {}

        pks = [(t.context, t.message_id) for t in translatables]
        pk_filter = tuple_(self.model.context, self.model.message_id).in_(pks)
        translations = self.session.query(self.model).filter(
            self.model.language == self.language).filter(pk_filter).values(
            self.model.context, self.model.message_id, self.model.value)
        cache = {(t[0], t[1]): t[2] for t in translations}
        return cache

    def _collect_translatables(self, translatable, collection=None):
        """
        Run over a translatable 'structure' and collect any translatables
        These are then bulk loaded from the db
        """

        if collection is None:
            collection = []

        if isinstance(translatable, TranslatableString):
            collection.append(translatable)
        elif isinstance(translatable, dict):
            [self._collect_translatables(val, collection)
                for val in translatable.itervalues()]
        elif isinstance(translatable, list):
            [self._collect_translatables(item, collection)
                for item in translatable]
        elif isinstance(translatable, tuple):
            [self._collect_translatables(item, collection)
                for item in translatable]

        return collection

    def save_translation(self, translatable, commit=True):
        if translatable.message_id is None:
            raise RuntimeError(
                "Cannot save translatable '{}'. "
                "Message id is None".format(translatable))

        if self.debug_output:
            debug_value = self._get_debug_translation(translatable)
            if translatable.pending_value == debug_value:
                return

        translation = self.model(
            context=translatable.context,
            message_id=translatable.message_id,
            language=self.language
        )

        # we can use merge for 'on duplicate key update'
        # (only works in sqla if we're matching on the primary key)
        translation = self.session.merge(translation)
        translation.value = translatable.pending_value

        if commit:
            self.session.commit()

    def delete_translations(self, translatable, commit=True):
        """ delete _all_ translations for this (context, message_id) """
        self.session.query(self.model).filter_by(
            context=translatable.context,
            message_id=translatable.message_id,
        ).delete()

        if commit:
            self.session.commit()

    def _normalised_translations(self, languages):
        """ helper for bulk operations

        returns query, aliases, columns, where

        ``query`` is a sqlalchemy query that will select all contexts
        and message_ids from the translations table and join it to itself
        once for each language supplied in the languages list, returning
        rows of the form (
            context,
            message_id,
            translation in language 1,
            translation in language 2,
            ...
        )

        ``columns`` is a list of all columns in this tuple

        ``aliases`` are sqlalchemy aliases to the (self) joined language tables
        so that we can apply filters, e.g.
            query.filter(aliases[0].value == NULL)

        """
        session = self.session
        model = self.model

        base_query = session.query(model.context, model.message_id).distinct()

        subquery = base_query.subquery(name='basequery')
        query = session.query(subquery)

        aliases = []
        for language in languages:
            alias = aliased(model, name=language)
            query = query.outerjoin(
                alias,
                and_(
                    alias.context == subquery.c.context,
                    alias.message_id == subquery.c.message_id,
                    alias.language == language
                )
            )
            aliases.append(alias)

        columns = [subquery.c.context, subquery.c.message_id] + [
            alias.value for alias in aliases]

        return query, aliases, columns

    def list_translations(self, languages):
        """ list all translations for the requested languages

        return a tuple (context, message_id, value1, value2, ...)

        where value_n is the translation for the nth language in the
        ``languages`` list
        """
        query, _, columns = self._normalised_translations(languages)
        return query.values(*columns)

    def list_missing_translations(self, languages):
        """ as ``list_translations`` but restricted to rows where a translation
        is missing for at least one of the requested languages
        """
        query, aliases, columns = self._normalised_translations(languages)
        query = query.filter(or_(*(alias.value == NULL for alias in aliases)))
        return query.values(*columns)


class TranslationContextManager(object):
    """ Knows all available ``message_id``\s for a given context """

    __metaclass__ = ABCMeta

    @abstractproperty
    def context(self):
        """ String used to identify translations managed by this manager """

    @abstractmethod
    def list_message_ids(self):
        """ List of message ids for all objects managed by this manager """

    # consider including _create_translation or something similar
    # from tests/models.py


class TranslationManager(object):
    """ Collection of ``TranslationContextManager``\s """

    def __init__(self):
        self._registry = {}

    def register(self, context_manager):
        context = context_manager.context
        if context in self._registry:
            raise KeyError(
                "ContextManager with context '{}' already registered".format(
                    context)
            )
        self._registry[context] = context_manager

    def list_contexts_and_message_ids(self, **kwargs):
        for context, context_manager_cls in self._registry.iteritems():
            context_manager = context_manager_cls(**kwargs)
            for message_id in context_manager.list_message_ids():
                yield (context, message_id)


translation_manager = TranslationManager()



import json

from taal import TranslatableString
from kaiso.types import get_type_id


def get_context(manager, obj, attribute_name):
    type_id = get_type_id(type(obj))
    return "taal:kaiso_field:{}:{}".format(type_id, attribute_name)


def get_message_id(manager, obj):
    primary_keys = list(manager.type_registry.get_index_entries(obj))
    return json.dumps(sorted(primary_keys))


def make_from_obj(manager, obj, attribute_name, pending_value=None):
    context = get_context(manager, obj, attribute_name)
    message_id = get_message_id(manager, obj)
    return TranslatableString(
        context=context,
        message_id=message_id,
        pending_value=pending_value
    )



from __future__ import absolute_import

from kaiso.attributes import String


class TranslatableString(String):
    @staticmethod
    def to_primitive(value, for_db):
        if not for_db:
            return value

        if value is not None:
            raise RuntimeError(
                "Cannot save directly to translated fields. "
                "Value was '{}'".format(value))



from weakref import WeakKeyDictionary

from kaiso.persistence import Manager as KaisoManager

from taal import TranslatableString as TaalTranslatableString
from taal.exceptions import NoTranslatorRegistered
from taal.kaiso import TranslatableString
from taal.kaiso.context_managers import (
    AttributeTranslationContextManager, TypeTranslationContextManager)
from taal.kaiso.types import get_context, get_message_id


translator_registry = WeakKeyDictionary()


def register_translator(owner, translator):
    translator_registry[owner] = translator


def get_translator(owner):
    try:
        return translator_registry[owner]
    except KeyError:
        raise NoTranslatorRegistered(
            "No translator registered for {}".format(owner))


def _label_attributes(type_id, attrs):
    labelled_attrs = []
    for attr in attrs:
        label = TaalTranslatableString(
            context=AttributeTranslationContextManager.context,
            message_id=AttributeTranslationContextManager.get_message_id(
                type_id, attr)
        )
        attr.label = label
        labelled_attrs.append(attr)
    return labelled_attrs


def collect_translatables(manager, obj):
    """ collect translatables from obj

        may also mutate obj to replace translations with placeholders

        returns an iterable yielding the collecting translatables or None

        expects translator.save_translation or translator.delete_translations
        to be called for each collected translatable
    """

    translations = []
    descriptor = manager.type_registry.get_descriptor(type(obj))
    for attr_name, attr_type in descriptor.attributes.items():
        attr = getattr(obj, attr_name)
        if isinstance(attr_type, TranslatableString):
            translations.append((attr_name, attr))
            # TODO: something better than this workaround
            # what do we want in the db?
            setattr(obj, attr_name, None)

    if not translations:
        # so that we can check if this is empty before we start iterating
        return None

    def iter_translatables():
        message_id = get_message_id(manager, obj)
        for attr_name, attr in translations:
            context = get_context(manager, obj, attr_name)
            translatable = TaalTranslatableString(
                context, message_id, attr)
            yield translatable

    return iter_translatables()


class Manager(KaisoManager):
    def serialize(self, obj):
        message_id = get_message_id(self, obj)
        data = super(Manager, self).serialize(obj)
        descriptor = self.type_registry.get_descriptor(type(obj))
        for attr_name, attr_type in descriptor.attributes.items():
            if isinstance(attr_type, TranslatableString):
                context = get_context(self, obj, attr_name)
                data[attr_name] = TaalTranslatableString(
                    context, message_id)
        return data

    def save(self, obj):
        translatables = collect_translatables(self, obj)
        saved = super(Manager, self).save(obj)

        if translatables is None:
            return saved

        translator = get_translator(self)

        for translatable in translatables:
            translator.save_translation(translatable)

        return saved

    def delete(self, obj):
        translatables = collect_translatables(self, obj)
        result = super(Manager, self).delete(obj)

        if translatables is None:
            return result

        translator = get_translator(self)

        for translatable in translatables:
            translator.delete_translations(translatable)

        return result

    def get_labeled_type_hierarchy(self, start_type_id=None):
        type_hierarchy = super(
            Manager, self).get_type_hierarchy(start_type_id)

        for type_id, bases, attrs in type_hierarchy:
            label = TaalTranslatableString(
                context=TypeTranslationContextManager.context,
                message_id=type_id
            )
            yield (
                type_id, label, bases, _label_attributes(type_id, attrs))



from __future__ import absolute_import

from sqlalchemy import Column, String, Text


class TranslationMixin(object):
    """ Mixin for sqlalchemy model to contain translation data

        Usage:
            Base = declarative_base()

            class MyTranslations(TranslationMixin, Base):
                __tablename__ = "my_translations"
    """

    context = Column(String(255, collation='utf8_bin'), primary_key=True)
    message_id = Column(String(255, collation='utf8_bin'), primary_key=True)
    language = Column(String(255, collation='utf8_bin'), primary_key=True)
    value = Column(Text(convert_unicode=True))



import json
from weakref import WeakSet

from sqlalchemy import event, inspect, types
from sqlalchemy.orm.mapper import Mapper

from taal import TranslatableString as TaalTranslatableString


CONTEXT_TEMPLATE = "taal:sa_field:{}:{}"
NOT_NULL = "taal:placeholder"


pending_translatables = WeakSet()  # to aid debugging


class NotNullValue(object):
    pass


class TranslatableString(types.TypeDecorator):

    impl = types.Text

    def process_bind_param(self, value, dialect):
        if value is None:
            return None

        if not isinstance(value, TaalTranslatableString):
            # this should only happen if someone is trying to query
            # TODO: verify this
            raise RuntimeError("Cannot filter on translated fields")

        if value in pending_translatables:
            pending_translatables.remove(value)
        else:
            raise RuntimeError(
                "Cannot save directly to translated fields. "
                "Use ``save_translation`` instead "
                "Value was '{}'".format(value))

        return NOT_NULL

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        elif value == NOT_NULL:
            # can't prevent this from being returned to the user
            # in the case of a direct query for Model.field
            # Return something that's more likely to error early
            # than a string
            return NotNullValue

        raise RuntimeError(
            "Unexpected value found in placeholer column: '{}'".format(
                value))


def get_context(obj, column):
    table = obj.__table__
    return CONTEXT_TEMPLATE.format(table, column)


def get_message_id(obj):
    cls = obj.__class__
    mapper = inspect(cls)
    primary_keys = mapper.primary_key_from_instance(obj)
    if any(key is None for key in primary_keys):
        return None
    return json.dumps(primary_keys)


def make_from_obj(obj, column, pending_value):
    context = get_context(obj, column)
    message_id = get_message_id(obj)

    if pending_value is NotNullValue:
        pending_value = None

    if isinstance(pending_value, TaalTranslatableString):
        raise TypeError("pending_value must be a string. Was '{}'".format(
            pending_value))

    return TaalTranslatableString(
        context=context,
        message_id=message_id,
        pending_value=pending_value
    )


@event.listens_for(Mapper, 'mapper_configured')
def register_listeners(mapper, cls):
    for column in mapper.columns:
        if isinstance(column.type, TranslatableString):
            from taal.sqlalchemy import events
            event.listen(cls, 'load', events.load)
            event.listen(cls, 'refresh', events.refresh)

            column_attr = getattr(cls, column.name)
            event.listen(column_attr, 'set', events.set_, retval=True)



from weakref import WeakKeyDictionary

from sqlalchemy import event, inspect

from taal import TranslatableString as TaalTranslatableString
from taal.sqlalchemy.types import (
    TranslatableString, pending_translatables, make_from_obj, NotNullValue)


translator_registry = WeakKeyDictionary()
flush_log = WeakKeyDictionary()


def register_translator(owner, translator):
    translator_registry[owner] = translator


def get_translator(owner):
    return translator_registry[owner]


def set_(target, value, oldvalue, initiator):
    """ Wrap any value in ``TranslatableString`` (except None) """
    if value is None:
        return None

    if isinstance(value, TaalTranslatableString):
        return TaalTranslatableString(
            value.context, value.message_id, value.pending_value)

    translatable = make_from_obj(target, initiator.key, value)
    return translatable


def load(target, context):
    """ Wrap columns when loading data from the db """
    mapper = inspect(target.__class__)
    for column in mapper.columns:
        if isinstance(column.type, TranslatableString):
            value = getattr(target, column.name)
            if value is None:
                continue
            elif value is NotNullValue:
                translatable = make_from_obj(target, column.name, value)
                setattr(target, column.name, translatable)
            elif isinstance(value, TaalTranslatableString):
                continue  # during session.merge
            else:
                raise TypeError("Unexpected column value '{}'".format(
                    value))


def refresh(target, args, attrs):
    mapper = inspect(target.__class__)
    if attrs is None:
        attrs = mapper.columns.keys()

    for column_name in attrs:
        if column_name not in mapper.columns:
            continue
        column = mapper.columns[column_name]
        if isinstance(column.type, TranslatableString):
            value = getattr(target, column.name)
            if value is not None:
                translatable = make_from_obj(target, column.name, value)
                setattr(target, column.name, translatable)
    return target


def add_to_flush_log(session, target, delete=False):
    mapper = inspect(target.__class__)
    for column in mapper.columns:
        if isinstance(column.type, TranslatableString):
            if delete:
                value = None  # will trigger deletion of translations
            else:
                value = getattr(target, column.name)
            if value is not None:
                pending_translatables.add(value)
                value = value.pending_value
            flush_log.setdefault(session, []).append(
                (session.transaction, target, column, value))


def before_flush(session, flush_context, instances):
    """ Queue up translations to be saved on commit. We don't want to
        save them until commit, but by then we no longer know what's
        changed. At this point we introspect the session to find added,
        changed and deleted objects and save them, along with information
        about the active session and transaction (to handle savepoint
        rollbacks) to the queue (``flush_log``)


        Notes:
            The ``instances`` argument is from a deprecated api
    """

    for target in session.dirty:
        if session.is_modified(target):
            add_to_flush_log(session, target)

    for target in session.new:
        add_to_flush_log(session, target)

    for target in session.deleted:
        add_to_flush_log(session, target, delete=True)


def after_bulk_update(session, query, query_context, result):
    # bulk updating to None would be ok, but leaves dangling Translations
    for bind in result.context.compiled.binds.values():
        field_type = bind.type
        if isinstance(field_type, TranslatableString):
            raise NotImplementedError("Bulk updates are not yet supported")


def after_commit(session):
    """ Save any pending translations for this session """
    for transaction, target, column, value in flush_log.pop(session, []):
        translator = get_translator(session)
        if value is None:
            # pending_value is ignored as we are deleting
            # just needs to be not-None
            translatable = make_from_obj(target, column.name, '')
            translator.delete_translations(translatable)
        else:
            translatable = make_from_obj(target, column.name, value)
            translator.save_translation(translatable, commit=True)

        old_value = getattr(target, column.name)
        if old_value is not None:
            # we may now have a primary key
            old_value.message_id = translatable.message_id
            # value is now saved. No need to keep around
            old_value.pending_value = None


def after_soft_rollback(session, previous_transaction):
    """ Drop any pending translations from this transaction """
    if session in flush_log:
        flush_log[session] = [
            pending for pending in flush_log[session]
            if pending[0] != previous_transaction]


def register_session(session):
    event.listen(session, 'before_flush', before_flush)
    event.listen(session, 'after_bulk_update', after_bulk_update)
    event.listen(session, 'after_commit', after_commit)
    event.listen(session, 'after_soft_rollback', after_soft_rollback)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
make_abs = lambda fn: os.path.join(here, fn)


def parse_requirments(fn, dependency_links):
    requirements = []

    if not os.path.exists(fn):
        return requirements, dependency_links

    with open(fn, 'r') as f:
        for dep in f:
            dep = dep.strip()
            # need to make github requirements work with
            # setuptools like it would work with `pip -r`
            # URLs will not work, so we transform them to
            # dependency_links and requirements
            if dep.startswith('git+'):
                dependency_links.append(dep)
                _, dep = dep.rsplit('#egg=', 1)
                dep = dep.replace('-', '==', 1)
            requirements.append(dep)

    return requirements, dependency_links


requirements, dependency_links = parse_requirments(
    make_abs('requirements.txt'), [])


setup(
    name='taal',
    packages=find_packages(exclude=['tests', 'tests.*']),
    version='0.1.1',
    author='onefinestay',
    author_email='engineering@onefinestay.com',
    url='https://github.com/onefinestay/taal',
    install_requires=requirements,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    description='Model translations',
    long_description=open(make_abs('README.rst')).read(),
    include_package_data=True,
    zip_safe=True,
)
